# RDF::Query
# -------------
# $Revision: 1.4 $
# $Date: 2004/02/25 19:56:04 $
# -----------------------------------------------------------------------------

=head1 NAME

RDF::Query - An RDQL implementation for RDF::Core

=cut

package RDF::Query;

=head1 SYNOPSIS

 my $query = new RDF::Query ( @files );
 my @rows = $query->query( <<'RDQL' );
   SELECT ?name
   WHERE (?person foaf:nick "kasei")
   USING foaf FOR <http://xmlns.com/foaf/0.1/>
 RDQL

=head1 DESCRIPTION

RDF::Query allows RDQL queries to be run against an RDF graph, returning rows
of matching results. See L<http://www.w3.org/Submission/2004/SUBM-RDQL-20040109/>
for more information.

=head1 REQUIRES

 L<Parse::RecDescent|Parse::RecDescent>
 L<RDF::Core|RDF::Core>
 L<LWP::Simple|LWP::Simple>

=cut


use strict;
use warnings;
use Carp qw(carp croak);

use Data::Dumper;
use Parse::RecDescent;
use LWP::Simple qw(get);

use RDF::Core::Model;
use RDF::Core::Query;
use RDF::Core::Model::Parser;
use RDF::Core::Storage::Memory;

######################################################################

our ($VERSION, $debug);
BEGIN {
	$::RD_TRACE	= undef;
	$::RD_HINT	= undef;
	$debug		= 0;
	$VERSION	= do { my @REV = split(/\./, (qw$Revision: 1.1 $)[1]); sprintf("%0.3f", $REV[0] + ($REV[1]/1000)) };
}

our($RDQL_GRAMMAR);
BEGIN {
	our $RDQL_GRAMMAR	= <<'END';
	query:			'SELECT' variable(s) SourceClause(?) 'WHERE' triplepattern(s) 'USING' namespaces
																	{
																		$return = {
																			variables	=> $item[2],
																			sources		=> $item[3][0],
																			triples		=> $item[5],
																			namespaces	=> $item[7]
																		}
																	}
	SourceClause:	('SOURCE' | 'FROM') Source(s)					{ $return = $item[2] }
	Source:			URI												{ $return = $item[1] }
	variable:		'?' identifier									{ $return = ['VAR',$item[2]] }
	triplepattern:	'(' variable URI VarUriConst ')'				{ $return = [ @item[2,3,4] ] }
	VarUriConst:	(variable | URI | CONST)						{ $return = $item[1] }
	namespaces:		namespace(s)									{ $return = { map { %{ $_ } } @{$item[1]} } }
	namespace:		identifier 'FOR' qURI							{ $return = {@item[1,3]} }
	identifier:		/(([a-zA-Z0-9_.-])+)/							{ $return = $1 }
	URI:			(qURI | QName)									{ $return = ['URI',$item[1]] }
	qURI:			'<' /[A-Za-z0-9_.!~*'()%;\/?:@&=+$,#-]+/ '>'	{ $return = $item[2] }
	QName:			identifier ':' /([^ \t<>()]+)/					{ $return = [@item[1,3]] }
	CONST:			(Number | Text)									{ $return = ['LITERAL',$item[1]] }
	Number:			/([0-9]+(\.[0-9]+)?)/							{ $return = $item[1] }
	Text:			dQText | sQText									{ $return = $item[1] }
	sQText:			"'" /([^']+)/ '"'								{ $return = $item[2] }
	dQText:			'"' /([^"]+)/ '"'								{ $return = $item[2] }
END
}

######################################################################

sub new {
	my $proto	= shift;
	my $class	= ref($proto) || $proto;
	my @files	= @_;
	my $rdql_parser	= new Parse::RecDescent ($RDQL_GRAMMAR);
	my $self 	= bless( { parser => $rdql_parser }, $class );
	$self->parse_files( @files );
	return $self;
}

sub query {
	my $self	= shift;
	my $query	= shift;
	my $model	= $self->model;
	my $parsed	= $self->parse_query( $query );
	local($self->{'parsed'})	= $parsed;
	warn "Parse tree: " . Dumper($self->parsed) if ($debug);
	print "Model contains " . $model->countStmts . " statement(s).\n" if ($debug);
	
	our $indent	= '';
	my @results	= $self->query_more( @{ $parsed->{'triples'} } );
	if (wantarray) {
		return @results;
	} else {
		my %results;
		@results{ @{ $parsed->{'variables'} } }	= @results;
		return \%results;
	}
}

sub query_more {
	my $self	= shift;
	my @triples	= @_;
	our $indent;
	
	my $parsed	= $self->parsed;
	my $model	= $self->model;
	my $triple	= shift(@triples);
	unless (ref($triple)) {
		carp "Something went wrong. No triple passed to query_more";
		return undef;
	}
	my @triple	= @{ $triple };
	warn "${indent}query_more: " . join(' ', map { UNIVERSAL::isa($_,'RDF::Core::Node') ? '<' . $_->getLabel . '>' : $_->[1] } @triple) . "\n" if ($debug);
	warn "${indent}-> with " . scalar(@triples) . " triples to go\n" if ($debug);
	if ($debug) {
		warn "${indent}-> more: " . join(' ', map { UNIVERSAL::isa($_,'RDF::Core::Node') ? '<' . $_->getLabel . '>' : $_->[1] } @{$_}) . "\n" for (@triples);
	}
	
	my $vars	= 0;
	my ($var, $method);
	
	my @methodmap	= ( 'getSubject', 'getPredicate', 'getObject' );
	for my $idx (0 .. 2) {
		my $data	= $triple[$idx];
		if (UNIVERSAL::isa($data, 'ARRAY') and $data->[0] eq 'VAR') {
			my $tmpvar	= $data->[1];
			my $val = $parsed->{'values'}{ $tmpvar };
			if (UNIVERSAL::isa($val, 'RDF::Core::Node')) {
				warn "${indent}-> already have value for $tmpvar: " . $val->getLabel . "\n" if ($debug);
				$triple[$idx]	= $val;
			} elsif (++$vars > 1) {
				warn "${indent}-> we've seen $vars variables in this triple... punt\n" if ($debug);
				push(@triples, $triple);
				return $self->query_more( @triples );
			} else {
				warn "${indent}-> found variable $var (we've seen $vars variables already)\n" if ($debug);
				$triple[$idx]	= undef;
				$var			= $tmpvar;
				$method			= $methodmap[ $idx ];
			}
		}
	}
	
	warn "${indent}getting $var\n" if ($debug);
	
	my @results;
	my $enum	= $model->getStmts( @triple );
	my $stmt	= $enum->getFirst;
	while (defined $stmt) {
		if ($vars) {
			warn "${indent}-> got variable $var = " . $stmt->$method()->getLabel . "\n" if ($debug);
			$parsed->{'values'}{ $var }	= $stmt->$method();
		}
		if (scalar(@triples)) {
			if ($debug) {
				warn "${indent}-> now for more triples...\n";
				warn "${indent}-> more: " . join(' ', map { UNIVERSAL::isa($_,'RDF::Core::Node') ? '<' . $_->getLabel . '>' : $_->[1] } @{$_}) . "\n" for (@triples);
			}
			$indent	.= '  ';
			push(@results, $self->query_more( @triples ));
		} else {
			warn "${indent}-> no triples left: result: " . $stmt->getLabel . "\n" if ($debug);
			push(@results, [ map { $parsed->{'values'}{$_} } map { $_->[1] } @{ $parsed->{'variables'} } ]);
		}
		$stmt = $enum->getNext;
		undef $parsed->{'values'}{ $var };
	}
	$enum->close;
	substr($indent, -2, 2)	= '';
	return @results;
}

sub parse_query {
	my $self	= shift;
	my $query	= shift;
	my $parser	= $self->parser;
	my $parsed	= $parser->query( $query );
	
	## CONVERT URIs to RDF::Core::Resources, and strings to RDF::Core::Literals
	foreach my $triple (@{ $parsed->{'triples'} }) {
		my ($sub,$pred,$obj)	= @{ $triple };
		my $preduri			= join('', $parsed->{'namespaces'}{$pred->[1][0]}, $pred->[1][1]);
		$triple->[1]	= new RDF::Core::Resource ($preduri);
		
		if (ref($obj)) {
			if ($obj->[0] eq 'LITERAL') {
				my $literal		= new RDF::Core::Literal ($obj->[1]);
				$triple->[2]	= $literal;
			} elsif ($obj->[0] eq 'URI') {
				my $resource	= ref($obj->[1])
								? join('', $parsed->{'namespaces'}{$obj->[1][0]}, $obj->[1][1])
								: $obj->[1];
				$triple->[2]	= new RDF::Core::Resource ($resource);
			}
		} else {
			warn "Object not a reference: " . Dumper($obj) . ' ';
		}
	}
	
	## LOAD ANY EXTERNAL RDF FILES
	my $sources	= $parsed->{'sources'};
	if (UNIVERSAL::isa( $sources, 'ARRAY' ) and scalar(@{ $sources })) {
		$self->parse_urls( @{ $sources } );
	}
	
	return $parsed;
}

sub parse_files {
	my $self	= shift;
	my @files	= @_;
	my $model	= $self->model;
	
	foreach my $file (@files) {
		warn "parsing $file\n" if ($debug);
		my %options = (
					Model		=> $model,
					Source		=> $file,
					SourceType	=> 'file',
					BaseURI		=> "http://kasei.us/e/ns/querybase",
					BNodePrefix	=> "genid"
				);
		my $parser	= new RDF::Core::Model::Parser (%options);
		$parser->parse;
	}
}

sub parse_urls {
	my $self	= shift;
	my @urls	= @_;
	my $model	= $self->model;
	
	foreach my $urldata (@urls) {
		next unless (UNIVERSAL::isa( $urldata, 'ARRAY' ) and $urldata->[0] eq 'URI');
		my $url	= $urldata->[1];
		my $rdf	= get($url);
		print "parsing $url\n" if ($debug);
		my %options = (
					Model		=> $model,
					Source		=> $rdf,
					SourceType	=> 'string',
					BaseURI		=> "http://kasei.us/e/ns/querybase",
					BNodePrefix	=> "genid"
				);
		my $parser	= new RDF::Core::Model::Parser (%options);
		$parser->parse;
	}
}

sub model {
	my $proto	= shift;
	my $model;
	if (ref($proto) and ref($proto->{'model'})) {
		$model	= $proto->{'model'};
	} else {
		my $storage	= new RDF::Core::Storage::Memory;
		$model	= new RDF::Core::Model (Storage => $storage);
		if (ref($proto)) {
			$proto->{'model'}	= $model;
		}
	}
	return $model;
}

sub AUTOLOAD {
	my $self	= $_[0];
	my $class	= ref($_[0]) || return undef;
	our $AUTOLOAD;
	return if ($AUTOLOAD =~ /DESTROY$/);
	my $method		= $AUTOLOAD;
	$method			=~ s/^.*://;
	
	if (substr($method,0,1) ne '_' and exists($self->{ $method })) {
		no strict 'refs';
		*$AUTOLOAD	= sub {
			my $self        = shift;
			my $class       = ref($self);
			return unless ($class);
			
			my $data	= $self->{ $method };
			if (@_) {
				$self->{ $method }	= shift;
			}
			return $data;
		};
		goto &$method;
	} else {
		croak qq[Can't locate object method "$method" via package $class];
	}
}


1;

__END__

=back

=head1 REVISION HISTORY

 $Log$
 
=head1 AUTHOR

 Gregory Williams <gwilliams@cpan.org>

=cut
